/* instruction.c */

/* Copyright (C) 2013 Niels Möller
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#if HAVE_CONFIG_H
# include "config.h"
#endif

#include <assert.h>

#include "instruction.h"

#include "xalloc.h"

/* Opcodes for various special cases */
#define OPCODE_LDST_INDEXED	0x4000
#define OPCODE_IMM_TST		0xa800
#define OPCODE_IMM_CMPEQ	0xaa00
#define OPCODE_IMM_CMPUGEQ	0xac00
#define OPCODE_IMM_CMPSGEQ	0xae00
#define OPCODE_XOR		0xe500

static struct instruction *
instruction_create (enum instruction_type type, enum keyword op, unsigned code,
		    unsigned align, unsigned size, unsigned lineno)
{
  NEW (struct instruction, instr);
  instr->lineno = lineno;
  instr->operation = op;
  instr->type = type;
  instr->code = code;
  instr->align = align;
  instr->size = size;
  return instr;
}

struct instruction *
instruction_int (enum keyword op, unsigned code,
		 uint64_t value, unsigned lineno)
{
  struct instruction *instr
    = instruction_create (INSTRUCTION_INT, op, code, code, code, lineno);

  instr->imm = value;
  return instr;
}

struct instruction *
instruction_nullary (enum keyword op, unsigned code,
		     unsigned lineno)
{
  return instruction_create (INSTRUCTION_NULLARY, op, code, 2, 2, lineno);
}

struct instruction *
instruction_unary (enum keyword op, unsigned code,
		   unsigned dreg, unsigned lineno)
{
  struct instruction *instr
    = instruction_create (INSTRUCTION_UNARY, op, code, 2, 2, lineno);
  instr->reg[0] = dreg;
  return instr;
}

struct instruction *
instruction_binary (enum keyword op, unsigned code,
		    unsigned dreg,
		    unsigned sreg, unsigned lineno)
{
  struct instruction *instr
    = instruction_create (INSTRUCTION_BINARY, op, code & ~1, 2, 2, lineno);
  if (code & 1)
    {
      /* Reversed args */
      instr->reg[0] = sreg;
      instr->reg[1] = dreg;
    }
  else
    {
      instr->reg[0] = dreg;
      instr->reg[1] = sreg;
    }
  return instr;
}

#define SWAP(a,b) do {				\
    unsigned swap_t = (a);			\
    (a) = (b);					\
    (b) = swap_t;				\
  }  while (0);

struct instruction *
instruction_ternary (enum keyword op, unsigned code,
		     unsigned dreg,
		     unsigned a0, unsigned a1, unsigned lineno)
{
  struct instruction *instr;

  /* Special case for indexed load and store */
  if (op == KEYWORD_LD)
    {
      code = OPCODE_LDST_INDEXED;
      if (a0 == a1)
	{
	  fprintf (stderr, "%d: Indexed load must use distinct base and index registers.\n",
		   lineno);
	  return NULL;
	}
      else if (a1 < a0)
	SWAP (a1, a0);
    }
  else if (op == KEYWORD_ST)
    {
      code = OPCODE_LDST_INDEXED;
      if (a0 == a1)
	{
	  fprintf (stderr, "%d: Indexed store must use distinct base and index registers.\n",
		   lineno);
	  return NULL;
	}
      else if (a1 > a0)
	SWAP (a1, a0);
    }

  instr = instruction_create (INSTRUCTION_TERNARY, op, code, 2, 2, lineno);
  instr->reg[0] = dreg;
  instr->reg[1] = a0;
  instr->reg[2] = a1;
  return instr;
}

struct instruction *
instruction_add (enum keyword op, unsigned code,
		 unsigned dreg,
		 unsigned sreg, int cin, unsigned lineno)
{
  if (cin)
    code |= 0x400;
  return instruction_binary (op, code, dreg, sreg, lineno);
}

/* Return 5-bit code, or -1 if not available */
static int
imm5_code (int64_t x)
{
  static const signed char tab[33] =    
    { /*  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 */
         -1, 0, 4, 1, 8, 2, 5, 3,12,-1, 6,-1, 9,-1, 7,-1,13,
      /*    17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 */
	 -1,-1,-1,10,-1,-1,-1,14,-1,-1,-1,11,-1,-1,-1,15,
    };

  if (x < -32 || x > 32)
    return -1;
  else if (x < 0)
    {
      int code = tab[-x];
      return code >= 0 ? (code | 16) : -1;
    }
  else
    return tab[x];
}

/* Check if x fits in an n-bit signed representation */
#define IMM_FIT(x, n) \
  ((x) >= - ((int64_t) 1 << (n-1)) && (x) < ((int64_t) 1 << (n-1)))

struct instruction *
instruction_imm (enum keyword op, unsigned code,
		 unsigned dreg,
		 int64_t imm, unsigned lineno)
{
  struct instruction *instr;
  unsigned size;
  if (!imm)
    switch (op)
      {
      case KEYWORD_MOV:
      case KEYWORD_AND:
	return instruction_binary (KEYWORD_XOR, OPCODE_XOR, dreg, dreg, lineno);
      case KEYWORD_CMPEQ:
	code = OPCODE_IMM_CMPUGEQ;
	imm = 1;
	break;
      case KEYWORD_CMPSGEQ:
	code = OPCODE_IMM_CMPUGEQ;
	imm = -1;
	break;
      default:
	fprintf (stderr, "%d: Questionable immediate zero argument.\n",
		 lineno);
      }
  if (op == KEYWORD_CMPUGEQ)
    {
      switch (imm)
	{
	case -1:
	  code = OPCODE_IMM_CMPEQ;
	  break;
	case 1:
	  code = OPCODE_IMM_TST;
	  imm = -1;
	  break;
	case 2:
	  code = OPCODE_IMM_TST;
	  imm = -2;
	  break;
	case 4:
	  code = OPCODE_IMM_TST;
	  imm = -4;
	  break;
	case 8:
	  code = OPCODE_IMM_TST;
	  imm = -8;
	  break;
	case 16:
	  code = OPCODE_IMM_TST;
	  imm = -16;
	  break;
	case 32:
	  code = OPCODE_IMM_TST;
	  imm = -32;
	  break;
	}
    }

  if (imm5_code (imm) >= 0)
    size = 2;
  else if (IMM_FIT (imm, 17))
    size = 4;
  else if (IMM_FIT (imm, 29))
    size = 6;
  else if (IMM_FIT (imm, 41))
    size = 8;
  else if (IMM_FIT (imm, 53))
    size = 10;
  else
    size = 12;

  instr = instruction_create (INSTRUCTION_IMM, op, code, 2, size, lineno);
  instr->reg[0] = dreg;
  instr->imm = imm;
  return instr;
}

struct instruction *
instruction_imm_add (enum keyword op, unsigned code,
		     unsigned dreg, int64_t imm, int cin,
		     unsigned lineno)
{
  if (imm == 0)
    {
      if (!cin)
	{
	  fprintf (stderr, "%d: Useless add/sub with zero.\n", lineno);
	  return NULL;
	}
      else
	{
	  fprintf (stderr, "%d: add , cc, #0 not yet implemented.\n",
		   lineno);
	  return NULL;
	}	
    }
  if (code & 1)
    /* FIXME: For subtract with carry, use ~imm (one's complement)
       rather than -imm? */
    imm = -imm;
  code &= ~1;
  if (cin)
    code |= 0x800;

  return instruction_imm (op, code, dreg, imm, lineno);
}

struct instruction *
instruction_imm_shift (enum keyword op, unsigned code,
		       unsigned dreg,
		       int64_t imm, unsigned lineno)
{
  if (imm <= 0 || imm >= 64)
    {
      fprintf (stderr, "%d: Shift count out of range.\n", lineno);
      return 0;
    }
  return instruction_unary (op, code | (imm << 4), dreg, lineno);
}


static int
imm_cmp_code (int64_t x)
{
  static const signed char tab[17] =
    { /*  10 12 14 16 18 20 22 24 26 28 30 32 */
           0, 1, 2, 3,-1, 4,-1, 5,-1,-6,-1, 7
    };
  if (x < -32 || x > 32)
    return -1;
  if (x < 10 && x > -10)
    return -1;
  if (x < 0)
    {
      int code = tab[(-x - 10)/2];
      return code >= 0 ? (code | 8) : -1;
    }
  else
    return tab[(x-10)/2];  
}

struct instruction *
instruction_imm_cmpgt (enum keyword op, unsigned code,
		       unsigned dreg,
		       int64_t imm, unsigned lineno)
{
  if (imm_cmp_code (imm) >= 0)
    {
      struct instruction *instr
	= instruction_create (INSTRUCTION_IMM_CMPGT,
			      op, code, 2, 2, lineno);
      instr->reg[0] = dreg;
      instr->imm = imm;

      return instr;
    }
  else if (op == KEYWORD_CMPUGT)
    switch (imm)
      {
      case -1:
	fprintf (stderr, "%d: Always false cmpugt.\n", lineno);
	return NULL;
      case 8:
	return instruction_imm (KEYWORD_CMPUGT,
				OPCODE_IMM_CMPUGEQ, dreg, -2, lineno);
      default:
	return instruction_imm (KEYWORD_CMPUGEQ, OPCODE_IMM_CMPUGEQ,
				dreg, imm+1, lineno);
      }
  else if (op == KEYWORD_CMPSGT)
    switch (imm)
      {
      case ((uint64_t) 1 << 63) - 1:
	fprintf (stderr, "%d: Always false cmpsgt.\n", lineno);
	return NULL;	
      case 0:
	return instruction_imm (KEYWORD_CMPSGT,
				OPCODE_IMM_CMPUGEQ, dreg, -1, lineno);
      case 8:
	return instruction_imm (KEYWORD_CMPSGT,
				OPCODE_IMM_CMPUGEQ, dreg, -4, lineno);
      default:
	return instruction_imm (KEYWORD_CMPSGEQ,
				OPCODE_IMM_CMPSGEQ, dreg, imm + 1, lineno);
      }
  else
    {
      fprintf (stderr, "%d: instruction_imm_cmpgt: Internal error.\n", lineno);
      return NULL;	
    }
}

static int
imm4_code (int64_t x)
{
  static const signed char tab[17] =
    { /*  0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 */
         -1, 0, 2, 1, 4,-1, 3,-1, 6,-1,-1,-1, 5,-1,-1,-1,7,
    };

  if (x < -16 || x > 16)
    return -1;
  else if (x < 0)
    {
      int code = tab[-x];
      return code >= 0 ? (code | 8) : -1;
    }
  else
    return tab[x];
}

struct instruction *
instruction_memory (enum keyword op, unsigned code, unsigned alt_code,
		    unsigned dreg, unsigned sreg,		    
		    int64_t offset, int update,
		    unsigned lineno)
{
  struct instruction *instr;
  unsigned size;

  if (!offset)
    return instruction_binary (op, alt_code, dreg, sreg, lineno);

  if (imm4_code (offset) >= 0)
    size = 2;
  else if (IMM_FIT (offset, 16))
    size = 4;
  else if (IMM_FIT (offset, 28))
    size = 6;
  else if (IMM_FIT (offset, 40))
    size = 8;
  else if (IMM_FIT (offset, 52))
    size = 10;
  else
    size = 12;

  if (update > 0 && offset < 0)
    {
      fprintf (stderr, "%d: Post-increment requires positive offset.\n",
	       lineno);
      return NULL;
    }

  if (update < 0 && offset > 0)
    {
      fprintf (stderr, "%d: Pre-decrement requires negative offset.\n",
	       lineno);
      return NULL;
    }

  if (update)
    code |= 0x1000;

  instr = instruction_create (INSTRUCTION_MEMORY, op, code, 2, size, lineno);
  instr->reg[0] = dreg;
  instr->reg[1] = sreg;
  instr->imm = offset;
  return instr;
}

struct instruction *
instruction_branch (enum keyword op, unsigned code,
		    const struct symbol *symbol, unsigned lineno)
{
  /* FIXME: Offset must fit without prefix. */
  struct instruction *instr
    = instruction_create (INSTRUCTION_BRANCH, op, code, 2, 2, lineno);
  instr->symbol = symbol;
  return instr;
}

struct instruction *
instruction_reg_branch (enum keyword op, unsigned code,
			unsigned sreg,
			unsigned lineno)
{
  if (code & 1)
    return instruction_binary (op, code & ~(uint64_t) 1, 15, sreg, lineno);
  else
    return instruction_unary (op, code, sreg, lineno);
}

static int
output_int (FILE *f, unsigned size, int64_t x)
{
  while (size-- > 0)
    if (fputc ((x >> (8*size)) & 0xff, f) < 0)
      return 0;

  return 1;
}

static int
output_uint16 (FILE *f, unsigned x)
{
  return output_int (f, 2, x);
}

static int
output_code_reg (FILE *f, unsigned code, unsigned reg)
{
  assert ((code & 0xf) == 0);
  assert (reg < 0x10);
  return output_uint16 (f, code | reg);
}

static int
output_code_2reg (FILE *f, unsigned code, unsigned dreg, unsigned sreg)
{
  assert ((code & 0xff) == 0);
  assert (dreg < 0x10);
  assert (sreg < 0x10);
  return output_uint16 (f, code | (sreg << 4) | dreg);
}

static int
output_code_3reg (FILE *f, unsigned code, unsigned dreg, unsigned a0, unsigned a1)
{
  assert ((code & 0xff) == 0);
  assert (dreg < 0x10);
  assert (a0 < 0x10);
  assert (a1 < 0x10);
  return output_uint16 (f, code | (a1 << 8) | (a0 << 4) | dreg);
}

static int
output_branch (FILE *f, const struct instruction *self)
{
  uint64_t target;

  assert (self->size == 2);
  assert ( (self->code & 0x3ff) == 0);

  if (!symbol_value (self->symbol, &target))
    {
      fprintf (stderr, "%d: Undefined label '%s'.\n",
	       self->lineno, self->symbol->name);
      return 0;
    }
  if (target & 1)
    {
      fprintf (stderr, "%d: Branch target '%s' refers to odd address %jx\n",
	       self->lineno, self->symbol->name, (uintmax_t) target);
      return 0;
    }
  int64_t offset = target/2 - (self->address + self->size)/2;
  if (!IMM_FIT (offset, 10))
    {
      fprintf (stderr, "%d: Branch target '%s' too far away: %jx\n",
	       self->lineno, self->symbol->name, (uintmax_t) target);
      return 0;
    }
  return output_uint16 (f, self->code | (offset & 0x3ff));  
}

static int
output_prefix (FILE *f, int64_t prefix, unsigned n)
{
  assert (n <= 5);
  assert ( (prefix >> (12*n)) == 0 || ( (~prefix) >> (12*n)) == 0);
  while (n-- > 0)
    if (!output_uint16 (f, 0x7000 | (0xfff & (prefix >> (12*n)))))
      return 0;
  return 1;  
}

static int
output_imm5 (FILE *f, unsigned code, unsigned dreg, unsigned imm)
{
  assert ((code & 0x1ff) == 0);
  assert (dreg < 0x10);
  assert (imm < 0x20);
  return output_uint16 (f, code | (imm << 4) | dreg);
}

static int
output_imm_cmpgt (FILE *f, unsigned code, unsigned dreg, unsigned imm)
{
  assert ((code & 0xff) == 0);
  assert (dreg < 0x10);
  assert (imm < 0x10);
  return output_uint16 (f, code | (imm << 4) | dreg);  
}

int
instruction_output (const struct instruction *self,
		    FILE *f)
{
  switch (self->type)
    {
    case INSTRUCTION_INT:
      return output_int (f, self->size, self->imm);
    case INSTRUCTION_NULLARY:
      return output_uint16 (f, self->code);
      break;
    case INSTRUCTION_UNARY:
      return output_code_reg (f, self->code, self->reg[0]);
      break;
    case INSTRUCTION_BINARY:
      return output_code_2reg (f, self->code, self->reg[0], self->reg[1]);
      break;
    case INSTRUCTION_TERNARY:
      return output_code_3reg (f, self->code, self->reg[0], self->reg[1], self->reg[2]);
    case INSTRUCTION_MEMORY:
      {
	int imm = imm4_code (self->imm);
	if (imm < 0)
	  {
	    if (!output_prefix (f, self->imm >> 4, self->size / 2 - 1))
	      return 0;
	    imm = self->imm & 15;
	  }
	return output_code_3reg (f, self->code, self->reg[0], self->reg[1], imm);
      }
    case INSTRUCTION_BRANCH:
      return output_branch (f, self);
    case INSTRUCTION_IMM:
      {
	int imm = imm5_code (self->imm);
	if (imm < 0)
	  {
	    if (!output_prefix (f, self->imm >> 5, self->size / 2 - 1))
	      return 0;
	    imm = self->imm & 31;
	  }
	return output_imm5 (f, self->code, self->reg[0], imm);
      }
    case INSTRUCTION_IMM_CMPGT:
      {
	int imm = imm5_code (self->imm);
	assert (imm >= 0);
	return output_imm_cmpgt (f, self->code, self->reg[0], imm);
      }
    default:
      fprintf (stderr, "instruction_output: type %d not implemented.\n",
	       self->type);
      return 0;
    }
}
