/*
    instr16 assembler
    Copyright (C) 2006, 2013  Niels Möller

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
%{
  #if HAVE_CONFIG_H
  # include "config.h"
  #endif

  #include <ctype.h>
  #include <stdarg.h>
  #include <stdint.h>
  #include <stdlib.h>

  #include "keyword.h"
  #include "program.h"
  #include "instruction.h"
  #include "symbol.h"

  struct parse_state
  {
    FILE *file;
    const char *filename;
    unsigned lineno;

    unsigned nerror;
    struct program *program;
  };

  static int
  yylex (struct parse_state *state);

  static void
  yyerror (struct parse_state *state, const char *msg);
%}

%union {
  int64_t number;
  const struct keyword_assoc *keyword;
  struct symbol *symbol;
  struct instruction *instruction;
}

%debug
%parse-param {struct parse_state *state}
%lex-param {struct parse_state *state}

%token TOK_NEWLINE
%token TOK_ERROR
%token <symbol> TOK_SYMBOL

%token <number> TOK_NUMBER

%token <keyword> TOK_SCALAR_TYPE
%token <keyword> TOK_INSTR_NULLARY
%token <keyword> TOK_INSTR_UNARY
%token <keyword> TOK_INSTR_BINARY
%token <keyword> TOK_INSTR_TERNARY
%token <keyword> TOK_INSTR_SHIFT
%token <keyword> TOK_INSTR_ADD
%token <keyword> TOK_INSTR_MEMORY
%token <keyword> TOK_INSTR_CMPGT
%token <keyword> TOK_INSTR_BRANCH
%token <keyword> TOK_REGISTER
%token <keyword> TOK_CC

%type <instruction> instruction
%type <instruction> instr_ld_st
%type <instruction> instr_add
%type <instruction> instr_branch
%type <symbol> label

%type <number> number

%%

program: /* Empty */
	| lines
	;

lines: line
	| lines line
	;

line: TOK_NEWLINE
	| instruction TOK_NEWLINE
	  {
	    if (!$1)
		YYERROR;
	    program_add_instruction (state->program, $1); }
	| label opt_newlines instruction TOK_NEWLINE
	  {
	    if (!$1)
		YYERROR;
	    program_add_instruction (state->program, $3);
	    program_define_label (state->program, $1, $3, state->lineno);
	  }
	;

/* FIXME: Attach line number and check for multiple definition here? */
label: TOK_SYMBOL ':' ; 

opt_newlines: /* Empty */
	| opt_newlines TOK_NEWLINE
	;

instruction: 	instr_ld_st
	|	instr_branch
	|	instr_add
	| 	TOK_SCALAR_TYPE number
		{ $$ = instruction_int ($1->id, $1->code, $2, state->lineno); }
	| 	TOK_INSTR_NULLARY
		{ $$ = instruction_nullary ($1->id, $1->code, state->lineno); }
	| 	TOK_INSTR_UNARY TOK_REGISTER
		{ $$ = instruction_unary ($1->id, $1->code, $2->code, state->lineno); }
	| 	TOK_INSTR_BINARY TOK_REGISTER ',' TOK_REGISTER
		{ $$ = instruction_binary ($1->id, $1->code, $2->code, $4->code,
		state->lineno); }
	|	TOK_INSTR_SHIFT TOK_REGISTER ',' '#' number
		{ $$ = instruction_imm_shift ($1->id, $1->code, $2->code, $5, state->lineno); }
	|	TOK_INSTR_SHIFT TOK_REGISTER ',' TOK_REGISTER
		{
		  if (!$1->alt)
		    {
		      fprintf (stderr,
			       "%d: Shift instruction %s requires immediate shift count.\n",
			       state->lineno, $1->name);
		      YYERROR;
		    }
		  $$ = instruction_binary ($1->id, $1->alt, $2->code,
					   $4->code, state->lineno);
		}
	|	TOK_INSTR_BINARY TOK_REGISTER ',' '#' number
		{
		  if (!$1->alt)
		    {
		      fprintf (stderr,
			       "%d: Instruction %s requires register arguments.\n",
			       state->lineno, $1->name);
		      YYERROR;
		    }
		  $$ = instruction_imm ($1->id, $1->alt, $2->code, $5, state->lineno);
		}
	|	TOK_INSTR_TERNARY TOK_REGISTER ',' TOK_REGISTER ',' TOK_REGISTER
	{ $$ = instruction_ternary ($1->id, $1->code, $2->code, $4->code, $6->code, state->lineno); }
	|	TOK_INSTR_CMPGT TOK_REGISTER ',' '#' number
		{ $$ = instruction_imm_cmpgt ($1->id, $1->code, $2->code, $5, state->lineno); }
		;

instr_ld_st:	TOK_INSTR_MEMORY TOK_REGISTER ',' '[' TOK_REGISTER ']'
		{
		  $$ = instruction_memory ($1->id, $1->code, $1->alt,
					   $2->code, $5->code, 0, 0, state->lineno);
		}
	|	TOK_INSTR_MEMORY TOK_REGISTER ',' '[' TOK_REGISTER ',' '#' number ']'
		{
		  $$ = instruction_memory ($1->id, $1->code, $1->alt,
					   $2->code, $5->code, $8, 0, state->lineno);
		}
	|	TOK_INSTR_MEMORY TOK_REGISTER ',' '[' TOK_REGISTER ',' '#' number ']' '!'
		{
		  $$ = instruction_memory ($1->id, $1->code, $1->alt,
					   $2->code, $5->code, $8, -1, state->lineno);
		}
	|	TOK_INSTR_MEMORY TOK_REGISTER ',' '[' TOK_REGISTER ']' ',' '#' number
		{
		  $$ = instruction_memory ($1->id, $1->code, $1->alt,
					   $2->code, $5->code, $9, 1, state->lineno);
		} 
	| 	TOK_INSTR_MEMORY TOK_REGISTER ',' '[' TOK_REGISTER ',' TOK_REGISTER ']'
		{ $$ = instruction_ternary ($1->id, $1->code, $2->code, $5->code,
		  $7->code, state->lineno);}
		;

instr_add:	TOK_INSTR_ADD TOK_REGISTER ',' TOK_REGISTER
		{ $$ = instruction_add ($1->id, $1->code, $2->code, $4->code, 0,
		state->lineno); }
	| 	TOK_INSTR_ADD TOK_REGISTER ',' TOK_CC ',' TOK_REGISTER
		{ $$ = instruction_add ($1->id, $1->code, $2->code, $6->code, 1,
		state->lineno); }
	| 	TOK_INSTR_ADD TOK_REGISTER ',' '#' number
		{ $$ = instruction_imm_add ($1->id, $1->alt, $2->code, $5, 0, state->lineno); }
	| 	TOK_INSTR_ADD TOK_REGISTER ',' TOK_CC ',' '#' number
		{ $$ = instruction_imm_add ($1->id, $1->alt, $2->code, $7, 1, state->lineno); }

	;

instr_branch:	TOK_INSTR_BRANCH TOK_SYMBOL
		{ $$ = instruction_branch ($1->id, $1->code, $2, state->lineno); }
	| 	TOK_INSTR_BRANCH '[' TOK_REGISTER ']'
		{ $$ = instruction_reg_branch ($1->id, $1->alt, $3->code, state->lineno); }
		;

/* Expressions (quite limited) */
number:	TOK_NUMBER
	| '-' number { $$ = -$2; }
	;
%%

static void
yyerror (struct parse_state *state, const char *msg)
{
  if (msg)
    fprintf (stderr, "%s:%d: %s.\n", state->filename, state->lineno, msg);
  state->nerror++;
}

static __attribute__ ((__format__ (__printf__, 2, 3))) int
lexical_error (struct parse_state *state, const char *msg, ...)
{
  va_list args;

  va_start (args, msg);

  fprintf (stderr, "%s:%d: ", state->filename, state->lineno);
  vfprintf (stderr, msg, args);
  fprintf (stderr, ".\n");

  va_end (args);
  return TOK_ERROR;
}

static int
skip_ws (FILE *f)
{
  int c;

  do
    c = getc(f);
  while (c != '\n' && isspace (c));

  return c;
}

static void
skip_comment (FILE *f)
{
  int c;

  do
    c = getc(f);
  while (c != EOF && c != '\n');
}

static int
parse_number(FILE *f, int64_t *value, int c)
{
  uint64_t x = c - '0';
  for (c = getc(f); isdigit(c); c = getc(f))
    x = 10*x + c - '0';

  ungetc(c, f);
  *value = x;

  return TOK_NUMBER;
}

static int
parse_char(FILE *f, int64_t *value)
{
  int c = getc (f);
  if (c == '\\')
    {
      c = getc (f);
      switch (c)
	{
	case '0': c = '\0'; break;
	case 'a': c = '\a'; break;
	case 'b': c = '\b'; break;
	case 'f': c = '\f'; break;
	case 'n': c = '\n'; break;
	case 'r': c = '\r'; break;
	case 't': c = '\t'; break;
	case 'v': c = '\v'; break;
	case '\'': break;
	case '\\' : break;
	default:
	  fprintf (stderr, "Bad character constant.\n");
	  return TOK_ERROR;
	}
      }
  *value = (unsigned char) c;

  if (getc (f) != '\'')
    {
      fprintf (stderr, "Bad character constant.\n");
      return TOK_ERROR;
    }
  

  return TOK_NUMBER;
}

#define MAX_ID_LENGTH 200

static int
parse_identifier(struct parse_state *state, YYSTYPE *value, int c)
{
  unsigned l;
  char buf[MAX_ID_LENGTH];
  const struct keyword_assoc *keyword;

  buf[0] = c;
  l = 1;

  for (;;)
    {
      c = getc (state->file);
      if (! (c == '_' || c == '.' ||
	     (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z')
	     || (c >= '0' && c <= '9')))
	break;

      if (l >= MAX_ID_LENGTH)
	return lexical_error(state, "Too long identifier");

      buf[l++] = c;
    }
  ungetc(c, state->file);

  keyword = keyword_lookup (buf, l);
  if (keyword)
    {
      value->keyword = keyword;
      return keyword->type;
    }
  else
    {
      value->symbol = symbol_intern (state->program->symbols,
				     buf, l, state->lineno);
      return TOK_SYMBOL;
    }
}

static int
yylex (struct parse_state *state)
{
  int c;

  c = skip_ws(state->file);

  switch (c)
    {
    case EOF:
      return -1;
    case ';':
      skip_comment(state->file);
      state->lineno++;
      return TOK_NEWLINE;
    case '\n':
      state->lineno++;
      return TOK_NEWLINE;
      /* FIXME: At the start of a line, treat # specially, comment or
	 line directive. */
    case ':':
    case ',':
    case '=':
    case '#':
    case '[':
    case ']':
    case '-':
    case '!':
      return c;
    case '\'':
      return parse_char (state->file, &yylval.number);
    default:
      if (isdigit(c))
	return parse_number(state->file, &yylval.number, c);
      else if (c == '_' || c == '.'
	       || (c >= 'a' && c <= 'z') || (c >= 'A' && c <= 'Z'))
	return parse_identifier(state, &yylval, c);
      else
	return TOK_ERROR;
    }
}

struct program *
program_parse_file (const char *filename, FILE *f)
{
  struct parse_state state;
  int res;

  state.file = f;
  state.filename = filename;
  state.lineno = 1;
  state.nerror = 0;

  state.program = program_create();

  if (getenv ("YYDEBUG"))
    yydebug = 1;

  res = yyparse (&state);

  if (res == 0 && state.nerror == 0)
    return state.program;
  else
    return NULL;
}
